@@ -1,11 +1,12 @@
 from abc import ABCMeta
 from enum import Enum
-from typing import Any, Tuple, Callable, Iterable
+from typing import Any, Tuple, Callable, Iterable, List
 
 from pypads import logger
 from pypads.app.base import tracking_active
 from pypads.utils.util import is_package_available
-from pypads import logger
+
+from pypads_padre.concepts.util import _tolist
 
 
 class Types(Enum):
@@ -67,6 +68,10 @@ class Crawler:
     def data(self):
         return self._data
 
+    @data.setter
+    def data(self, data):
+        self._data = data
+
     @property
     def format(self):
         return self._format
@@ -102,7 +107,8 @@ class Crawler:
             self._use_args = True
         else:
             for _ctx in self._modules:
-                if _ctx.value == self._callback.__module__ or _ctx.value in self._callback.__module__ or self._callback.__module__ in _ctx.value:
+                if _ctx.value == self._callback.__module__ or _ctx.value in self._callback.__module__ or \
+                        self._callback.__module__ in _ctx.value:
                     self._format = _ctx.value
                     self._fn = self._fns.get(_ctx.value, self._fn)
                     self._use_args = True
@@ -144,26 +150,32 @@ class Crawler:
         return obj.data, metadata, targets
 
 
+# TODO feature metadata extraction (type: [categorical, continuous,..] and statistics [range, freq, ...])
 # --- Numpy array object ---
-def numpy_crawler(obj: Crawler, **kwargs):
+def numpy_crawler(obj: Crawler, target_columns=None, **kwargs):
     logger.info("Detecting a dataset object of type 'numpy.ndarray'. Crawling any available metadata...")
-    features = [(str(i), str(obj.data[:, i].dtype), False, (obj.data[:, i].min(), obj.data[:, i].max())) for i in
-                range(obj.data.shape[1])]
+    # , (obj.data[:, i].min(), obj.data[:, i].max())
+    if len(obj.data.shape) == 2:
+        features = [(str(i), str(obj.data[:, i].dtype), False) for i in
+                    range(obj.data.shape[1])]
+    else:
+        # TODO for multidim datasets
+        features = None
     metadata = {"type": str(obj.format), "shape": obj.data.shape, "features": features}
     metadata = {**metadata, **kwargs}
     targets = None
     try:
-        targets_col = None
-        for k in kwargs:
-            if isinstance(kwargs.get(k), Iterable):
-                if "target" in k or "class" in k:
-                    targets_col = kwargs.get(k)
-                    break
-        if targets_col:
-            targets = obj.data[:, targets_col]
-            metadata["features"][targets_col][2] = True
+        if target_columns:
+            targets = obj.data[:, target_columns]
+            if isinstance(target_columns, Iterable):
+                for c in target_columns:
+                    feature = metadata["features"][c]
+                    metadata["features"][c] = (feature[0], feature[1], True)
+            else:
+                feature = metadata["features"][target_columns]
+                metadata["features"][target_columns] = (feature[0], feature[1], True)
     except Exception as e:
-        print(str(e))
+        logger.warning(str(e))
     return obj.data, metadata, targets
 
 
@@ -171,19 +183,20 @@ Crawler.register_fn(Types.ndarray.value, numpy_crawler)
 
 
 # --- Pandas Dataframe object ---
-def dataframe_crawler(obj: Crawler, **kwargs):
+def dataframe_crawler(obj: Crawler, target_columns, **kwargs):
     logger.info("Detecting a dataset object of type 'pandas.DataFrame'. Crawling any available metadata...")
     data = obj.data
     features = []
     for i, col in enumerate(data.columns):
-        features.append((col, str(data[col].dtype), False, (data[col].min(), data[col].max())))
-    metadata = {"type": obj.format, "shape": data.shape, "features": features}
+        flag = col in target_columns if target_columns is not None else False
+        features.append((col, str(data[col].dtype), flag))
+    metadata = {"type": str(obj.format), "shape": data.shape, "features": features}
     metadata = {**metadata, **kwargs}
-    if "target" in data.columns:
-        targets = data[[col for col in data.columns if "target" in col]].values
+    targets = None
+    if target_columns is not None:
+        targets = data[target_columns].values
     else:
-        logger.warning("Target values might be innaccurate.")
-        targets = data[[data.columns[-1]]].values
+        logger.warning("Target values might be innaccurate or not tracked.")
     return data, metadata, targets
 
 
@@ -194,7 +207,7 @@ Crawler.register_fn(Types.dataframe.value, dataframe_crawler)
 def series_crawler(obj: Crawler, **kwargs):
     logger.info("Detecting a dataset object of type 'pandas.Series'. Crawling any available metadata...")
     data = obj.data
-    metadata = {"type": obj.format, "shape": data.shape}
+    metadata = {"type": str(obj.format), "shape": data.shape}
     metadata = {**metadata, **kwargs}
     return data, metadata, None
 
@@ -206,9 +219,10 @@ def bunch_crawler(obj: Crawler, **kwargs):
     data = np.concatenate([bunch.get('data'), bunch.get("target").reshape(len(bunch.get("target")), 1)], axis=1)
     features = []
     for i, name in enumerate(bunch.get("feature_names")):
-        features.append((name, str(data[:, i].dtype), False, (data[:, i].min(), data[:, i].max())))
-    features.append(("class", str(data[:, -1].dtype), True, tuple(bunch.get("target_names"))))
-    metadata = {"type": str(obj.format), "features": features, "description": bunch.get("DESCR"), "shape": data.shape}
+        features.append((name, str(data[:, i].dtype), False))
+    features.append(("class", str(data[:, -1].dtype), True))
+    metadata = {"type": str(obj.format), "features": features, "classes": _tolist(bunch.get("target_names")),
+                "description": bunch.get("DESCR"), "shape": data.shape}
     metadata = {**metadata, **kwargs}
     return data, metadata, bunch.get("target")
 
@@ -219,9 +233,9 @@ def sklearn_crawler(obj: Crawler, **kwargs):
     if "return_X_y" in kwargs and kwargs.get("return_X_y"):
         X, y = obj.data
         data = np.concatenate([X, y.reshape(len(y), 1)], axis=1)
-        features = [(str(i), str(X[:, i].dtype), False, (data[:, i].min(), data[:, i].max())) for i in
+        features = [(str(i), str(X[:, i].dtype), False) for i in
                     range(X.shape[1])]
-        features.append(("class", str(y.dtype), True, (y.min(), y.max())))
+        features.append(("class", str(y.dtype), True))
         metadata = {"type": str(obj.format), "features": features, "shape": (X.shape[0], X.shape[1] + 1)}
         metadata = {**metadata, **kwargs}
         return data, metadata, y
@@ -234,13 +248,15 @@ Crawler.register_fn(Modules.sklearn.value, sklearn_crawler)
 
 
 # --- TorchVision Dataset object ---
-def torch_crawler(obj: Crawler, *args, **kwargs):
+def torch_crawler(obj: Crawler, **kwargs):
     logger.info("Detecting a torchvision dataset loaded object. Crawling any available metadata...")
     data = obj.data.data.numpy()
     targets = obj.data.targets.numpy()
+    train = obj.data.train
+    source = obj.data.training_file if train else obj.data.test_file
     metadata = {"format": obj.format, "shape": data.shape, "classes": obj.data.classes,
-                "Description": obj.data.__repr__()}
-    metadata = {**metadata, **kwargs}
+                "Description": obj.data.__repr__(), "training_data": train, "source": source}
+    # metadata = {**metadata, **kwargs}
     return data, metadata, targets
 
 
@@ -249,7 +265,7 @@ if is_package_available("torchvision"):
 
 
 # --- Keras datasets ---
-def keras_crawler(obj: Crawler, *args, **kwargs):
+def keras_crawler(obj: Crawler, **kwargs):
     logger.info("Detecting a keras dataset loaded object. Crawling any available metadata...")
     (X_train, y_train), (X_test, y_test) = obj.data
     import numpy as np
@@ -274,3 +290,39 @@ def graph_crawler(obj: Crawler, **kwargs):
 
 
 Crawler.register_fn(Types.graph.value, graph_crawler)
+
+
+# --- tuple returned dataset
+def tuple_crawler(obj: Crawler, output_format, **kwargs):
+    crawler = obj
+    data = obj.data
+    metadata = dict()
+    data_outputs = dict()
+    targets = None
+    if output_format is None:
+        # do primitive crawling
+        for i, o in enumerate(data):
+            crawler.data = o
+            crawler._identify_data_object()
+            d_o, meta, _ = crawler.crawl(**{})
+            data_outputs["part_{}".format(i)] = d_o
+            metadata["part_{}".format(i)] = meta
+        metadata = {**metadata, **kwargs}
+        return data_outputs, metadata, targets
+    else:
+        for i, (key, value) in enumerate(output_format.items()):
+            if value == "features":
+                crawler.data = data[i]
+                crawler._identify_data_object()
+                d_o, meta, _ = crawler.crawl(**{})
+                metadata[key] = meta
+                data_outputs[key] = d_o
+            elif value == "targets" or value == "labels":
+                metadata[key] = {"type": str(obj.format), "targets_shape": data[i].shape}
+                data_outputs[key] = data[i]
+                targets = data[i]
+        metadata = {**metadata, **kwargs}
+        return data_outputs, metadata, targets
+
+
+Crawler.register_fn(Types.tuple.value, tuple_crawler)
